@@ -1,18 +1,5 @@
 #include "server.h"
 
-// client message
-#define INPUT '0'
-#define PING '1'
-#define RESIZE_TERMINAL '2'
-#define JSON_DATA '{'
-
-// server message
-#define OUTPUT '0'
-#define PONG '1'
-#define SET_WINDOW_TITLE '2'
-#define SET_PREFERENCES '3'
-#define SET_RECONNECT '4'
-
 #define BUF_SIZE 1024
 
 int
@@ -42,30 +29,29 @@ send_initial_message(struct lws *wsi) {
     return 0;
 }
 
-struct winsize *
-parse_window_size(const char *json) {
+bool
+parse_window_size(const char *json, struct winsize *size) {
     int columns, rows;
     json_object *obj = json_tokener_parse(json);
     struct json_object *o = NULL;
 
     if (!json_object_object_get_ex(obj, "columns", &o)) {
         lwsl_err("columns field not exists!\n");
-        return NULL;
+        return false;
     }
     columns = json_object_get_int(o);
     if (!json_object_object_get_ex(obj, "rows", &o)) {
         lwsl_err("rows field not exists!\n");
-        return NULL;
+        return false;
     }
     rows = json_object_get_int(o);
     json_object_put(obj);
 
-    struct winsize *size = t_malloc(sizeof(struct winsize));
     memset(size, 0, sizeof(struct winsize));
     size->ws_col = (unsigned short) columns;
     size->ws_row = (unsigned short) rows;
 
-    return size;
+    return true;
 }
 
 bool
@@ -93,14 +79,12 @@ check_host_origin(struct lws *wsi) {
 
 void
 tty_client_destroy(struct tty_client *client) {
-    if (client->exit || client->pid <= 0)
+    if (!client->running || client->pid <= 0)
         return;
-
-    // stop event loop
-    client->exit = true;
+    client->running = false;
 
     // kill process and free resource
-    lwsl_notice("sending %s to process %d\n", server->sig_name, client->pid);
+    lwsl_notice("sending %s (%d) to process %d\n", server->sig_name, server->sig_code, client->pid);
     if (kill(client->pid, server->sig_code) != 0) {
         lwsl_err("kill: pid, errno: %d (%s)\n", client->pid, errno, strerror(errno));
     }
@@ -112,9 +96,9 @@ tty_client_destroy(struct tty_client *client) {
 
     // free the buffer
     if (client->buffer != NULL)
-        t_free(client->buffer);
+        free(client->buffer);
 
-    // remove from clients list
+    // remove from client list
     pthread_mutex_lock(&server->lock);
     LIST_REMOVE(client, list);
     server->client_count--;
@@ -150,22 +134,24 @@ thread_run_command(void *args) {
             lwsl_notice("started process, pid: %d\n", pid);
             client->pid = pid;
             client->pty = pty;
+            client->running = true;
+            if (client->size.ws_row > 0 && client->size.ws_col > 0)
+                ioctl(client->pty, TIOCSWINSZ, &client->size);
 
-            while (!client->exit) {
+            while (client->running) {
                 FD_ZERO (&des_set);
                 FD_SET (pty, &des_set);
 
-                if (select(pty + 1, &des_set, NULL, NULL, NULL) < 0) {
+                if (select(pty + 1, &des_set, NULL, NULL, NULL) < 0)
                     break;
-                }
 
                 if (FD_ISSET (pty, &des_set)) {
                     memset(buf, 0, BUF_SIZE);
                     bytes = (int) read(pty, buf, BUF_SIZE);
-                    struct pty_data *frame = (struct pty_data *) t_malloc(sizeof(struct pty_data));
+                    struct pty_data *frame = (struct pty_data *) xmalloc(sizeof(struct pty_data));
                     frame->len = bytes;
                     if (bytes > 0) {
-                        frame->data = t_malloc((size_t) bytes);
+                        frame->data = xmalloc((size_t) bytes);
                         memcpy(frame->data, buf, bytes);
                     }
                     pthread_mutex_lock(&client->lock);
@@ -173,7 +159,6 @@ thread_run_command(void *args) {
                     pthread_mutex_unlock(&client->lock);
                 }
             }
-            tty_client_destroy(client);
             break;
     }
 
@@ -184,21 +169,27 @@ int
 callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
              void *user, void *in, size_t len) {
     struct tty_client *client = (struct tty_client *) user;
-    struct winsize *size;
+    char buf[256];
 
     switch (reason) {
         case LWS_CALLBACK_FILTER_PROTOCOL_CONNECTION:
             if (server->once && server->client_count > 0) {
-                lwsl_notice("refuse to serve new client due to the --once option.\n");
-                return -1;
+                lwsl_warn("refuse to serve WS client due to the --once option.\n");
+                return 1;
             }
+            if (lws_hdr_copy(wsi, buf, sizeof(buf), WSI_TOKEN_GET_URI) <= 0 || strcmp(buf, WS_PATH)) {
+                lwsl_warn("refuse to serve WS client for illegal ws path: %s\n", buf);
+                return 1;
+            }
+
             if (server->check_origin && !check_host_origin(wsi)) {
-                lwsl_notice("refuse to serve new client from different origin due to the --check-origin option.\n");
-                return -1;
+                lwsl_warn("refuse to serve WS client from different origin due to the --check-origin option.\n");
+                return 1;
             }
             break;
+
         case LWS_CALLBACK_ESTABLISHED:
-            client->exit = false;
+            client->running = false;
             client->initialized = false;
             client->authenticated = false;
             client->wsi = wsi;
@@ -207,17 +198,14 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
                                    client->hostname, sizeof(client->hostname),
                                    client->address, sizeof(client->address));
             STAILQ_INIT(&client->queue);
-            if (pthread_create(&client->thread, NULL, thread_run_command, client) != 0) {
-                lwsl_err("pthread_create\n");
-                return -1;
-            }
 
             pthread_mutex_lock(&server->lock);
             LIST_INSERT_HEAD(&server->clients, client, list);
             server->client_count++;
             pthread_mutex_unlock(&server->lock);
+            lws_hdr_copy(wsi, buf, sizeof(buf), WSI_TOKEN_GET_URI);
 
-            lwsl_notice("client connected from %s (%s), total: %d\n", client->hostname, client->address, server->client_count);
+            lwsl_notice("WS   %s - %s (%s), clients: %d\n", buf, client->address, client->hostname, server->client_count);
             break;
 
         case LWS_CALLBACK_SERVER_WRITEABLE:
@@ -234,7 +222,7 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
                 // read error or client exited, close connection
                 if (frame->len <= 0) {
                     STAILQ_REMOVE_HEAD(&client->queue, list);
-                    t_free(frame);
+                    free(frame);
                     return -1;
                 }
 
@@ -244,7 +232,7 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
                 unsigned char *p = &message[LWS_PRE];
                 size_t n = sprintf((char *) p, "%c%s", OUTPUT, b64_text);
 
-                t_free(b64_text);
+                free(b64_text);
 
                 if (lws_write(wsi, p, n, LWS_WRITE_TEXT) < n) {
                     lwsl_err("lws_write\n");
@@ -252,10 +240,10 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
                 }
 
                 STAILQ_REMOVE_HEAD(&client->queue, list);
-                t_free(frame->data);
-                t_free(frame);
+                free(frame->data);
+                free(frame);
 
-                if(lws_partial_buffered(wsi)){
+                if (lws_partial_buffered(wsi)) {
                     lws_callback_on_writable(wsi);
                     break;
                 }
@@ -265,11 +253,11 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
 
         case LWS_CALLBACK_RECEIVE:
             if (client->buffer == NULL) {
-                client->buffer = t_malloc(len + 1);
+                client->buffer = xmalloc(len + 1);
                 client->len = len;
                 memcpy(client->buffer, in, len);
             } else {
-                client->buffer = t_realloc(client->buffer, client->len + len + 1);
+                client->buffer = xrealloc(client->buffer, client->len + len + 1);
                 memcpy(client->buffer + client->len, in, len);
                 client->len += len;
             }
@@ -279,8 +267,8 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
 
             // check auth
             if (server->credential != NULL && !client->authenticated && command != JSON_DATA) {
-                lwsl_notice("websocket authentication failed\n");
-                return -1;
+                lwsl_warn("WS client not authenticated\n");
+                return 1;
             }
 
             // check if there are more fragmented messages
@@ -290,6 +278,8 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
 
             switch (command) {
                 case INPUT:
+                    if (client->pty == 0)
+                        break;
                     if (server->readonly)
                         return 0;
                     if (write(client->pty, client->buffer + 1, client->len - 1) < client->len - 1) {
@@ -307,44 +297,46 @@ callback_tty(struct lws *wsi, enum lws_callback_reasons reason,
                     }
                     break;
                 case RESIZE_TERMINAL:
-                    size = parse_window_size(client->buffer + 1);
-                    if (size != NULL) {
-                        if (ioctl(client->pty, TIOCSWINSZ, size) == -1) {
+                    if (parse_window_size(client->buffer + 1, &client->size) && client->pty > 0) {
+                        if (ioctl(client->pty, TIOCSWINSZ, &client->size) == -1) {
                             lwsl_err("ioctl TIOCSWINSZ: %d (%s)\n", errno, strerror(errno));
                         }
-                        t_free(size);
                     }
                     break;
                 case JSON_DATA:
-                    if (server->credential == NULL)
+                    if (client->pid > 0)
                         break;
-                    {
+                    if (server->credential != NULL) {
                         json_object *obj = json_tokener_parse(client->buffer);
                         struct json_object *o = NULL;
                         if (json_object_object_get_ex(obj, "AuthToken", &o)) {
                             const char *token = json_object_get_string(o);
-                            if (strcmp(token, server->credential)) {
-                                lwsl_notice("websocket authentication failed with token: %s\n", token);
-                                return -1;
+                            if (token == NULL || strcmp(token, server->credential)) {
+                                lwsl_warn("WS authentication failed with token: %s\n", token);
+                                return 1;
                             }
                         }
                         client->authenticated = true;
                     }
+                    if (pthread_create(&client->thread, NULL, thread_run_command, client) != 0) {
+                        lwsl_err("pthread_create\n");
+                        return 1;
+                    }
                     break;
                 default:
-                    lwsl_notice("unknown message type: %c\n", command);
+                    lwsl_warn("unknown message type: %c\n", command);
                     break;
             }
 
             if (client->buffer != NULL) {
-                t_free(client->buffer);
+                free(client->buffer);
                 client->buffer = NULL;
             }
             break;
 
         case LWS_CALLBACK_CLOSED:
             tty_client_destroy(client);
-            lwsl_notice("client disconnected from %s (%s), total: %d\n", client->hostname, client->address, server->client_count);
+            lwsl_notice("WS closed from %s (%s), clients: %d\n", client->address, client->hostname, server->client_count);
             if (server->once && server->client_count == 0) {
                 lwsl_notice("exiting due to the --once option.\n");
                 force_exit = true;
